@@ -48,7 +48,7 @@ usage(FILE *stream, const char *progname)
 	fprintf(stream, "\t-a\t\tfallback to EDNS0 and TCP if the answer is truncated\n");
 	fprintf(stream, "\t-b <bufsize>\tuse <bufsize> as the buffer size (defaults to 512 b)\n");
 	fprintf(stream, "\t-c <file>\t\tuse file for rescursive nameserver configuration (/etc/resolv.conf)\n");
-	fprintf(stream, "\t-k <file>\tspecify a file that contains a trusted DNSSEC key [**]\n");
+	fprintf(stream, "\t-k <file>\tspecify a file that contains a trusted DNSSEC key (DNSKEY|DS) [**]\n");
 	fprintf(stream, "\t\t\tused to verify any signatures in the current answer\n");
 	fprintf(stream, "\t-o <mnemonic>\tset flags to: [QR|qr][AA|aa][TC|tc][RD|rd][CD|cd][RA|ra][AD|ad]\n");
 	fprintf(stream, "\t\t\tlowercase: unset bit, uppercase: set bit\n");
@@ -339,6 +339,7 @@ main(int argc, char *argv[])
 				PURPOSE = DRILL_REVERSE;
 				break;
 			case 'y':
+#ifdef HAVE_SSL
 				if (strchr(optarg, ':')) {
 					tsig_separator = (size_t) (strchr(optarg, ':') - optarg);
 					if (strchr(optarg + tsig_separator + 1, ':')) {
@@ -360,6 +361,11 @@ main(int argc, char *argv[])
 					tsig_name[tsig_separator] = '\0';
 					tsig_data[ tsig_separator2 - tsig_separator - 1] = '\0';
 				}
+#else
+				fprintf(stderr, "TSIG requested, but SSL is not supported\n");
+				result = EXIT_FAILURE;
+				goto exit;
+#endif /* HAVE_SSL */
 				break;
 			case 'z':
 				qrandom = false;
@@ -461,8 +467,9 @@ main(int argc, char *argv[])
 		/* no server given make a resolver from /etc/resolv.conf */
 		status = ldns_resolver_new_frm_file(&res, resolv_conf_file);
 		if (status != LDNS_STATUS_OK) {
-			warning("Could not create a resolver structure: %s",
-				   ldns_get_errorstr_by_id(status));
+			warning("Could not create a resolver structure: %s (%s)\n"
+					"Try drill @localhost if you have a resolver running on your machine.",
+				    ldns_get_errorstr_by_id(status), resolv_conf_file);
 			result = EXIT_FAILURE;
 			goto exit;
 		}
@@ -773,7 +780,7 @@ main(int argc, char *argv[])
 					axfr_rr = ldns_axfr_next(res);
 					if(!axfr_rr) {
 						fprintf(stderr, "AXFR failed.\n");
-						ldns_pkt_print(stdout, 
+						ldns_pkt_print(stdout,
 							ldns_axfr_last_pkt(res));
 						goto exit;
 					}
@@ -905,9 +912,7 @@ main(int argc, char *argv[])
 	ldns_rr_list_deep_free(cmdline_rr_list);
 	ldns_rdf_deep_free(trace_start_name);
 	xfree(progname);
-/*
 	xfree(tsig_name);
-*/
 	xfree(tsig_data);
 	xfree(tsig_algorithm);
 
