@@ -3,7 +3,7 @@
  * It traces the authoritatives nameservers down from the root
  * And uses TCP, to minimize spoofing danger.
  *
- * (c) NLnet Labs, 2006
+ * (c) NLnet Labs, 2006 - 2008
  * See the file LICENSE for the license
  */
 
@@ -16,7 +16,7 @@ int verbosity = 0;
 uint8_t address_family = 0;
 bool store_in_file = false;
 
-void
+static void
 usage(FILE *fp, char *prog) {
 	fprintf(fp, "%s domain\n", prog);
 	fprintf(fp, "  retrieve the dnskeys for a domain\n");
@@ -24,12 +24,13 @@ usage(FILE *fp, char *prog) {
 	fprintf(fp, "-4\t\tUse IPv4 only\n");
 	fprintf(fp, "-6\t\tUse IPv6 only\n");
 	fprintf(fp, "-h\t\tShow this help\n");
+	fprintf(fp, "-i\t\tInsecurer mode; don't do checks, just query for the keys\n");
 	fprintf(fp, "-r <file>\tUse file to read root hints from\n");
 	fprintf(fp, "-s\t\tDon't print the keys but store them in files\n\t\tcalled K<file>.+<alg>.+<keytag>.key\n");
 	fprintf(fp, "-v <int>\tVerbosity level (0-5, not verbose-very verbose)\n");
 }
 
-ldns_rr_list *
+static ldns_rr_list *
 retrieve_dnskeys(ldns_resolver *local_res, ldns_rdf *name, ldns_rr_type t,
 		ldns_rr_class c, ldns_rr_list *dns_root)
 {
@@ -237,11 +238,12 @@ retrieve_dnskeys(ldns_resolver *local_res, ldns_rdf *name, ldns_rr_type t,
 						printf("Comparing answer list of answer to previous\n\n");
 					}
 					ldns_rr_list_sort(ldns_pkt_answer(p));
+					ldns_rr_list_sort(answer_list);
 					if (ldns_rr_list_compare(answer_list, ldns_pkt_answer(p)) != 0) {
 						fprintf(stderr, "ERROR: different answer answer from nameserver\n");
-						fprintf(stderr, "I had (from previous servers):\n");
+						fprintf(stderr, "\nI had (from previous servers):\n");
 						ldns_rr_list_print(stderr, answer_list);
-						fprintf(stderr, "I received (from nameserver at ");
+						fprintf(stderr, "\nI received (from nameserver at ");
 						ldns_rdf_print(stderr, ldns_resolver_nameservers(res)[0]);
 						fprintf(stderr, "):\n");
 						ldns_rr_list_print(stderr, ldns_pkt_answer(p));
@@ -259,11 +261,12 @@ retrieve_dnskeys(ldns_resolver *local_res, ldns_rdf *name, ldns_rr_type t,
 						printf("Comparing authority list of answer to previous\n\n");
 					}
 					ldns_rr_list_sort(ldns_pkt_authority(p));
+					ldns_rr_list_sort(authority_list);
 					if (ldns_rr_list_compare(authority_list, ldns_pkt_authority(p)) != 0) {
 						fprintf(stderr, "ERROR: different authority answer from nameserver\n");
-						fprintf(stderr, "I had (from previous servers):\n");
+						fprintf(stderr, "\nI had (from previous servers):\n");
 						ldns_rr_list_print(stderr, authority_list);
-						fprintf(stderr, "I received (from nameserver at ");
+						fprintf(stderr, "\nI received (from nameserver at ");
 						ldns_rdf_print(stderr, ldns_resolver_nameservers(res)[0]);
 						fprintf(stderr, "):\n");
 						ldns_rr_list_print(stderr, ldns_pkt_authority(p));
@@ -324,11 +327,12 @@ retrieve_dnskeys(ldns_resolver *local_res, ldns_rdf *name, ldns_rr_type t,
 						printf("Comparing answer list of answer to previous\n\n");
 					}
 					ldns_rr_list_sort(ldns_pkt_answer(p));
+					ldns_rr_list_sort(answer_list);
 					if (ldns_rr_list_compare(answer_list, ldns_pkt_answer(p)) != 0) {
 						fprintf(stderr, "ERROR: different answer answer from nameserver\n");
-						fprintf(stderr, "I had (from previous servers):\n");
+						fprintf(stderr, "\nI had (from previous servers):\n");
 						ldns_rr_list_print(stderr, answer_list);
-						fprintf(stderr, "I received (from nameserver at ");
+						fprintf(stderr, "\nI received (from nameserver at ");
 						ldns_rdf_print(stderr, ldns_resolver_nameservers(res)[0]);
 						fprintf(stderr, "):\n");
 						ldns_rr_list_print(stderr, ldns_pkt_answer(p));
@@ -346,11 +350,12 @@ retrieve_dnskeys(ldns_resolver *local_res, ldns_rdf *name, ldns_rr_type t,
 						printf("Comparing authority list of answer to previous\n\n");
 					}
 					ldns_rr_list_sort(ldns_pkt_authority(p));
+					ldns_rr_list_sort(authority_list);
 					if (ldns_rr_list_compare(authority_list, ldns_pkt_authority(p)) != 0) {
 						fprintf(stderr, "ERROR: different authority answer from nameserver\n");
-						fprintf(stderr, "I had (from previous servers):\n");
+						fprintf(stderr, "\nI had (from previous servers):\n");
 						ldns_rr_list_print(stderr, authority_list);
-						fprintf(stderr, "I received (from nameserver at ");
+						fprintf(stderr, "\nI received (from nameserver at ");
 						ldns_rdf_print(stderr, ldns_resolver_nameservers(res)[0]);
 						fprintf(stderr, "):\n");
 						ldns_rr_list_print(stderr, ldns_pkt_authority(p));
@@ -465,11 +470,12 @@ retrieve_dnskeys(ldns_resolver *local_res, ldns_rdf *name, ldns_rr_type t,
 				printf("1Comparing answer rr list of answer to previous\n");
 			}
 			ldns_rr_list_sort(ldns_pkt_answer(p));
+			ldns_rr_list_sort(answer_list);
 			if (ldns_rr_list_compare(answer_list, ldns_pkt_answer(p)) != 0) {
 				printf("ERROR: different answer section in response from nameserver\n");
-				fprintf(stderr, "I had:\n");
+				fprintf(stderr, "\nI had:\n");
 				ldns_rr_list_print(stderr, answer_list);
-				fprintf(stderr, "I received (from nameserver at ");
+				fprintf(stderr, "\nI received (from nameserver at ");
 				ldns_rdf_print(stderr, ldns_resolver_nameservers(res)[0]);
 				fprintf(stderr, "):\n");
 				ldns_rr_list_print(stderr, ldns_pkt_answer(p));
@@ -564,6 +570,9 @@ main(int argc, char *argv[])
 	ldns_buffer *outputfile_buffer;
 	FILE *outputfile;
 	ldns_rr *k;
+	
+	bool insecure = false;
+	ldns_pkt *pkt;
 
 	domain = NULL;
 	res = NULL;
@@ -588,6 +597,8 @@ main(int argc, char *argv[])
 			} else if (strncmp("-h", argv[i], 3) == 0) {
 				usage(stdout, argv[0]);
 				exit(EXIT_SUCCESS);
+			} else if (strncmp("-i", argv[i], 2) == 0) {
+				insecure = true;
 			} else if (strncmp("-r", argv[i], 2) == 0) {
 				if (strlen(argv[i]) > 2) {
 					root_file = argv[i]+2;
@@ -652,8 +663,13 @@ main(int argc, char *argv[])
 
 	ldns_resolver_set_ip6(res, address_family);
 
-	l = retrieve_dnskeys(res, domain, LDNS_RR_TYPE_DNSKEY, LDNS_RR_CLASS_IN, dns_root);
-
+	if (insecure) {
+		pkt = ldns_resolver_query(res, domain, LDNS_RR_TYPE_DNSKEY, LDNS_RR_CLASS_IN, LDNS_RD);
+		l = ldns_pkt_rr_list_by_type(pkt, LDNS_RR_TYPE_DNSKEY, LDNS_SECTION_ANY_NOQUESTION);
+	} else {
+		l = retrieve_dnskeys(res, domain, LDNS_RR_TYPE_DNSKEY, LDNS_RR_CLASS_IN, dns_root);
+	}
+	
 	/* separator for result data and verbosity data */
 	if (verbosity > 0) {
 		fprintf(stdout, "; ---------------------------\n");
