@@ -3,7 +3,7 @@
  *
  * Data storage for specially crafted replies for testing purposes.
  *
- * (c) NLnet Labs, 2005, 2006, 2007
+ * (c) NLnet Labs, 2005, 2006, 2007, 2008
  * See the file LICENSE for the license
  */
 
@@ -31,8 +31,12 @@ struct sockaddr_storage;
 /** string to show in warnings and errors */
 static const char* prog_name = "ldns-testpkts";
 
-/** logging routine, provided by caller. */
-void verbose(int lvl, const char* msg, ...);
+#ifndef UTIL_LOG_H
+/** verbosity definition for compat */
+enum verbosity_value { NO_VERBOSE=0 };
+#endif
+/** logging routine, provided by caller */
+void verbose(enum verbosity_value lvl, const char* msg, ...) ATTR_FORMAT(printf, 2, 3);
 
 /** print error and exit */
 static void error(const char* msg, ...)
@@ -61,14 +65,14 @@ static bool isendline(char c)
  * @param keyword: the keyword to match
  * @return: true if keyword present. False otherwise, and str unchanged.
 */
-static bool str_keyword(const char** str, const char* keyword)
+static bool str_keyword(char** str, const char* keyword)
 {
 	size_t len = strlen(keyword);
 	assert(str && keyword);
 	if(strncmp(*str, keyword, len) != 0)
 		return false;
 	*str += len;
-	while(isspace(**str))
+	while(isspace((int)**str))
 		(*str)++;
 	return true;
 }
@@ -92,9 +96,9 @@ entry_add_reply(struct entry* entry)
 }
 
 /** parse MATCH line */
-static void matchline(const char* line, struct entry* e)
+static void matchline(char* line, struct entry* e)
 {
-	const char* parse = line;
+	char* parse = line;
 	while(*parse) {
 		if(isendline(*parse)) 
 			return;
@@ -104,8 +108,16 @@ static void matchline(const char* line, struct entry* e)
 			e->match_qtype = true;
 		} else if(str_keyword(&parse, "qname")) {
 			e->match_qname = true;
+		} else if(str_keyword(&parse, "subdomain")) {
+			e->match_subdomain = true;
 		} else if(str_keyword(&parse, "all")) {
 			e->match_all = true;
+		} else if(str_keyword(&parse, "ttl")) {
+			e->match_ttl = true;
+		} else if(str_keyword(&parse, "DO")) {
+			e->match_do = true;
+		} else if(str_keyword(&parse, "noedns")) {
+			e->match_noedns = true;
 		} else if(str_keyword(&parse, "UDP")) {
 			e->match_transport = transport_udp;
 		} else if(str_keyword(&parse, "TCP")) {
@@ -116,7 +128,7 @@ static void matchline(const char* line, struct entry* e)
 				error("expected = or : in MATCH: %s", line);
 			parse++;
 			e->ixfr_soa_serial = (uint32_t)strtol(parse, (char**)&parse, 10);
-			while(isspace(*parse)) 
+			while(isspace((int)*parse)) 
 				parse++;
 		} else {
 			error("could not parse MATCH: '%s'", parse);
@@ -125,9 +137,9 @@ static void matchline(const char* line, struct entry* e)
 }
 
 /** parse REPLY line */
-static void replyline(const char* line, ldns_pkt *reply)
+static void replyline(char* line, ldns_pkt *reply)
 {
-	const char* parse = line;
+	char* parse = line;
 	while(*parse) {
 		if(isendline(*parse)) 
 			return;
@@ -153,6 +165,8 @@ static void replyline(const char* line, ldns_pkt *reply)
 			ldns_pkt_set_rcode(reply, LDNS_RCODE_NXDOMAIN);
 		} else if(str_keyword(&parse, "NOTIMPL")) {
 			ldns_pkt_set_rcode(reply, LDNS_RCODE_NOTIMPL);
+		} else if(str_keyword(&parse, "REFUSED")) {
+			ldns_pkt_set_rcode(reply, LDNS_RCODE_REFUSED);
 		} else if(str_keyword(&parse, "YXDOMAIN")) {
 			ldns_pkt_set_rcode(reply, LDNS_RCODE_YXDOMAIN);
 		} else if(str_keyword(&parse, "YXRRSET")) {
@@ -178,6 +192,9 @@ static void replyline(const char* line, ldns_pkt *reply)
 			ldns_pkt_set_ra(reply, true);
 		} else if(str_keyword(&parse, "AD")) {
 			ldns_pkt_set_ad(reply, true);
+		} else if(str_keyword(&parse, "DO")) {
+			ldns_pkt_set_edns_udp_size(reply, 4096);
+			ldns_pkt_set_edns_do(reply, true);
 		} else {
 			error("could not parse REPLY: '%s'", parse);
 		}
@@ -185,22 +202,24 @@ static void replyline(const char* line, ldns_pkt *reply)
 }
 
 /** parse ADJUST line */
-static void adjustline(const char* line, struct entry* e, 
+static void adjustline(char* line, struct entry* e, 
 	struct reply_packet* pkt)
 {
-	const char* parse = line;
+	char* parse = line;
 	while(*parse) {
 		if(isendline(*parse)) 
 			return;
 		if(str_keyword(&parse, "copy_id")) {
 			e->copy_id = true;
+		} else if(str_keyword(&parse, "copy_query")) {
+			e->copy_query = true;
 		} else if(str_keyword(&parse, "sleep=")) {
 			e->sleeptime = (unsigned int) strtol(parse, (char**)&parse, 10);
-			while(isspace(*parse)) 
+			while(isspace((int)*parse)) 
 				parse++;
 		} else if(str_keyword(&parse, "packet_sleep=")) {
 			pkt->packet_sleep = (unsigned int) strtol(parse, (char**)&parse, 10);
-			while(isspace(*parse)) 
+			while(isspace((int)*parse)) 
 				parse++;
 		} else {
 			error("could not parse ADJUST: '%s'", parse);
@@ -216,12 +235,17 @@ static struct entry* new_entry()
 	e->match_opcode = false;
 	e->match_qtype = false;
 	e->match_qname = false;
+	e->match_subdomain = false;
 	e->match_all = false;
+	e->match_ttl = false;
+	e->match_do = false;
+	e->match_noedns = false;
 	e->match_serial = false;
 	e->ixfr_soa_serial = 0;
 	e->match_transport = transport_any;
 	e->reply_list = NULL;
 	e->copy_id = false;
+	e->copy_query = false;
 	e->sleeptime = 0;
 	e->next = NULL;
 	return e;
@@ -377,7 +401,7 @@ get_origin(const char* name, int lineno, ldns_rdf** origin, char* parse)
 	*origin = NULL;
 
 	end=parse;
-	while(!isspace(*end) && !isendline(*end))
+	while(!isspace((int)*end) && !isendline(*end))
 		end++;
 	store = *end;
 	*end = 0;
@@ -391,12 +415,12 @@ get_origin(const char* name, int lineno, ldns_rdf** origin, char* parse)
 
 /* Reads one entry from file. Returns entry or NULL on error. */
 struct entry*
-read_entry(FILE* in, const char* name, int *lineno, uint16_t* default_ttl, 
+read_entry(FILE* in, const char* name, int *lineno, uint32_t* default_ttl, 
 	ldns_rdf** origin, ldns_rdf** prev_rr)
 {
 	struct entry* current = NULL;
 	char line[MAX_LINE];
-	const char* parse;
+	char* parse;
 	ldns_pkt_section add_section = LDNS_SECTION_QUESTION;
 	struct reply_packet *cur_reply = NULL;
 	bool reading_hex = false;
@@ -407,7 +431,7 @@ read_entry(FILE* in, const char* name, int *lineno, uint16_t* default_ttl,
 		parse = line;
 		(*lineno) ++;
 		
-		while(isspace(*parse))
+		while(isspace((int)*parse))
 			parse++;
 		/* test for keywords */
 		if(isendline(*parse))
@@ -422,10 +446,10 @@ read_entry(FILE* in, const char* name, int *lineno, uint16_t* default_ttl,
 			cur_reply = entry_add_reply(current);
 			continue;
 		} else if(str_keyword(&parse, "$ORIGIN")) {
-			get_origin(name, *lineno, origin, (char*)parse);
+			get_origin(name, *lineno, origin, parse);
 			continue;
 		} else if(str_keyword(&parse, "$TTL")) {
-			*default_ttl = (uint16_t)atoi(parse);
+			*default_ttl = (uint32_t)atoi(parse);
 			continue;
 		}
 
@@ -470,9 +494,12 @@ read_entry(FILE* in, const char* name, int *lineno, uint16_t* default_ttl,
 			/* it must be a RR, parse and add to packet. */
 			ldns_rr* n = NULL;
 			ldns_status status;
-			status = ldns_rr_new_frm_str(&n, parse, *default_ttl, 
-				*origin, prev_rr);
-			if (status != LDNS_STATUS_OK)
+			if(add_section == LDNS_SECTION_QUESTION)
+				status = ldns_rr_new_question_frm_str(
+					&n, parse, *origin, prev_rr);
+			else status = ldns_rr_new_frm_str(&n, parse, 
+				*default_ttl, *origin, prev_rr);
+			if(status != LDNS_STATUS_OK)
 				error("%s line %d:\n\t%s: %s", name, *lineno,
 					ldns_get_errorstr_by_id(status), parse);
 			ldns_pkt_push_rr(cur_reply->reply, add_section, n);
@@ -490,7 +517,7 @@ read_entry(FILE* in, const char* name, int *lineno, uint16_t* default_ttl,
 	return 0;
 }
 
-/** reads the canned reply file and returns a list of structs */
+/* reads the canned reply file and returns a list of structs */
 struct entry* 
 read_datafile(const char* name)
 {
@@ -499,7 +526,7 @@ read_datafile(const char* name)
 	struct entry* current = NULL;
 	FILE *in;
 	int lineno = 0;
-	uint16_t default_ttl = 0;
+	uint32_t default_ttl = 0;
 	ldns_rdf* origin = NULL;
 	ldns_rdf* prev_rr = NULL;
 	int entry_num = 0;
@@ -520,6 +547,8 @@ read_datafile(const char* name)
 	verbose(1, "%s: Read %d entries\n", prog_name, entry_num);
 
 	fclose(in);
+	ldns_rdf_deep_free(origin);
+	ldns_rdf_deep_free(prev_rr);
 	return list;
 }
 
@@ -555,7 +584,7 @@ static uint32_t get_serial(ldns_pkt* p)
 
 /** match two rr lists */
 static int
-match_list(ldns_rr_list* q, ldns_rr_list *p)
+match_list(ldns_rr_list* q, ldns_rr_list *p, bool mttl)
 {
 	size_t i;
 	if(ldns_rr_list_rr_count(q) != ldns_rr_list_rr_count(p))
@@ -564,7 +593,12 @@ match_list(ldns_rr_list* q, ldns_rr_list *p)
 	{
 		if(ldns_rr_compare(ldns_rr_list_rr(q, i), 
 			ldns_rr_list_rr(p, i)) != 0) {
-			verbose(3, "rr %d different", i);
+			verbose(3, "rr %d different", (int)i);
+			return 0;
+		}
+		if(mttl && ldns_rr_ttl(ldns_rr_list_rr(q, i)) !=
+			ldns_rr_ttl(ldns_rr_list_rr(p, i))) {
+			verbose(3, "rr %d ttl different", (int)i);
 			return 0;
 		}
 	}
@@ -583,7 +617,7 @@ cmp_bool(int x, int y)
 
 /** match all of the packet */
 static int
-match_all(ldns_pkt* q, ldns_pkt* p)
+match_all(ldns_pkt* q, ldns_pkt* p, bool mttl)
 {
 	if(ldns_pkt_get_opcode(q) != ldns_pkt_get_opcode(p)) 
 	{ verbose(3, "allmatch: opcode different"); return 0;}
@@ -613,18 +647,18 @@ match_all(ldns_pkt* q, ldns_pkt* p)
 	{ verbose(3, "allmatch: nscount different"); return 0;}
 	if(ldns_pkt_arcount(q) != ldns_pkt_arcount(p))
 	{ verbose(3, "allmatch: arcount different"); return 0;}
-	if(!match_list(ldns_pkt_question(q), ldns_pkt_question(p)))
+	if(!match_list(ldns_pkt_question(q), ldns_pkt_question(p), 0))
 	{ verbose(3, "allmatch: qd section different"); return 0;}
-	if(!match_list(ldns_pkt_answer(q), ldns_pkt_answer(p)))
+	if(!match_list(ldns_pkt_answer(q), ldns_pkt_answer(p), mttl))
 	{ verbose(3, "allmatch: an section different"); return 0;}
-	if(!match_list(ldns_pkt_authority(q), ldns_pkt_authority(p)))
+	if(!match_list(ldns_pkt_authority(q), ldns_pkt_authority(p), mttl))
 	{ verbose(3, "allmatch: ar section different"); return 0;}
-	if(!match_list(ldns_pkt_additional(q), ldns_pkt_additional(p))) 
+	if(!match_list(ldns_pkt_additional(q), ldns_pkt_additional(p), mttl)) 
 	{ verbose(3, "allmatch: ns section different"); return 0;}
 	return 1;
 }
 
-/** finds entry in list, or returns NULL */
+/* finds entry in list, or returns NULL */
 struct entry* 
 find_match(struct entry* entries, ldns_pkt* query_pkt,
 	enum transport_type transport)
@@ -651,15 +685,34 @@ find_match(struct entry* entries, ldns_pkt* query_pkt,
 				continue;
 			}
 		}
+		if(p->match_subdomain) {
+			if(!get_owner(query_pkt) || !get_owner(reply) ||
+				(ldns_dname_compare(get_owner(query_pkt), 
+				get_owner(reply)) != 0 &&
+				!ldns_dname_is_subdomain(
+				get_owner(query_pkt), get_owner(reply))))
+			{
+				verbose(3, "bad subdomain\n");
+				continue;
+			}
+		}
 		if(p->match_serial && get_serial(query_pkt) != p->ixfr_soa_serial) {
 				verbose(3, "bad serial\n");
 				continue;
 		}
+		if(p->match_do && !ldns_pkt_edns_do(query_pkt)) {
+			verbose(3, "no DO bit set\n");
+			continue;
+		}
+		if(p->match_noedns && ldns_pkt_edns(query_pkt)) {
+			verbose(3, "bad; EDNS OPT present\n");
+			continue;
+		}
 		if(p->match_transport != transport_any && p->match_transport != transport) {
 			verbose(3, "bad transport\n");
 			continue;
 		}
-		if(p->match_all && !match_all(query_pkt, reply)) {
+		if(p->match_all && !match_all(query_pkt, reply, p->match_ttl)) {
 			verbose(3, "bad allmatch\n");
 			continue;
 		}
@@ -675,9 +728,19 @@ adjust_packet(struct entry* match, ldns_pkt* answer_pkt, ldns_pkt* query_pkt)
 	/* copy & adjust packet */
 	if(match->copy_id)
 		ldns_pkt_set_id(answer_pkt, ldns_pkt_id(query_pkt));
+	if(match->copy_query) {
+		ldns_rr_list* list = ldns_pkt_get_section_clone(query_pkt,
+			LDNS_SECTION_QUESTION);
+		ldns_rr_list_deep_free(ldns_pkt_question(answer_pkt));
+		ldns_pkt_set_question(answer_pkt, list);
+	}
 	if(match->sleeptime > 0) {
 		verbose(3, "sleeping for %d seconds\n", match->sleeptime);
+#ifdef HAVE_SLEEP
 		sleep(match->sleeptime);
+#else
+		Sleep(match->sleeptime * 1000);
+#endif
 	}
 }
 
@@ -709,7 +772,7 @@ handle_query(uint8_t* inbuf, ssize_t inlen, struct entry* entries, int* count,
 	
 	query_rr = ldns_rr_list_rr(ldns_pkt_question(query_pkt), 0);
 	verbose(1, "query %d: id %d: %s %d bytes: ", ++(*count), (int)ldns_pkt_id(query_pkt), 
-		(transport==transport_tcp)?"TCP":"UDP", inlen);
+		(transport==transport_tcp)?"TCP":"UDP", (int)inlen);
 	if(verbose_out) ldns_rr_print(verbose_out, query_rr);
 	if(verbose_out) ldns_pkt_print(verbose_out, query_pkt);
 
@@ -750,9 +813,14 @@ handle_query(uint8_t* inbuf, ssize_t inlen, struct entry* entries, int* count,
 				answer_pkt = NULL;
 			} else {
 				verbose(3, "Could not parse hex data (%s), sending hex data directly.\n", ldns_get_errorstr_by_id(status));
+				/* still try to adjust ID */
 				answer_size = ldns_buffer_capacity(p->reply_from_hex);
 				outbuf = LDNS_XMALLOC(uint8_t, answer_size);
 				memcpy(outbuf, ldns_buffer_export(p->reply_from_hex), answer_size);
+				if(entry->copy_id) {
+					ldns_write_uint16(outbuf, 
+						ldns_pkt_id(query_pkt));
+				}
 			}
 		} else {
 			answer_pkt = ldns_pkt_clone(p->reply);
@@ -772,7 +840,11 @@ handle_query(uint8_t* inbuf, ssize_t inlen, struct entry* entries, int* count,
 		if(p->packet_sleep) {
 			verbose(3, "sleeping for next packet %d secs\n", 
 				p->packet_sleep);
+#ifdef HAVE_SLEEP
 			sleep(p->packet_sleep);
+#else
+			Sleep(p->packet_sleep * 1000);
+#endif
 			verbose(3, "wakeup for next packet "
 				"(slept %d secs)\n", p->packet_sleep);
 		}
