@@ -29,7 +29,7 @@ ldns_lookup_table ldns_signing_algorithms[] = {
         { LDNS_SIGN_RSASHA512, "RSASHA512" },
 #endif
 #ifdef USE_GOST
-        { LDNS_SIGN_GOST, "GOST" },
+        { LDNS_SIGN_ECC_GOST, "ECC-GOST" },
 #endif
         { LDNS_SIGN_DSA, "DSA" },
         { LDNS_SIGN_DSA_NSEC3, "DSA_NSEC3" },
@@ -76,7 +76,6 @@ ldns_key_new()
 		ldns_key_set_external_key(newkey, NULL);
 		return newkey;
 	}
-	return NULL;
 }
 
 ldns_status
@@ -104,6 +103,9 @@ ldns_key_new_frm_engine(ldns_key **key, ENGINE *e, char *key_id, ldns_algorithm
 #endif
 
 #ifdef USE_GOST
+/** store GOST engine reference loaded into OpenSSL library */
+ENGINE* ldns_gost_engine = NULL;
+
 int
 ldns_key_EVP_load_gost_id(void)
 {
@@ -133,26 +135,39 @@ ldns_key_EVP_load_gost_id(void)
 		return 0;
 	}
 	if(!ENGINE_set_default(e, ENGINE_METHOD_ALL)) {
+		ENGINE_finish(e);
 		ENGINE_free(e);
 		return 0;
 	}
 
 	meth = EVP_PKEY_asn1_find_str(&e, "gost2001", -1);
-	ENGINE_free(e);
 	if(!meth) {
 		/* algo not found */
+		ENGINE_finish(e);
+		ENGINE_free(e);
 		return 0;
 	}
+        /* Note: do not ENGINE_finish and ENGINE_free the acquired engine
+         * on some platforms this frees up the meth and unloads gost stuff */
+        ldns_gost_engine = e;
 	
 	EVP_PKEY_asn1_get0_info(&gost_id, NULL, NULL, NULL, NULL, meth);
 	return gost_id;
+} 
+
+void ldns_key_EVP_unload_gost(void)
+{
+        if(ldns_gost_engine) {
+                ENGINE_finish(ldns_gost_engine);
+                ENGINE_free(ldns_gost_engine);
+                ldns_gost_engine = NULL;
+        }
 }
 
 /** read GOST private key */
 static EVP_PKEY*
 ldns_key_new_frm_fp_gost_l(FILE* fp, int* line_nr)
 {
-	ssize_t len;
 	char token[16384];
 	const unsigned char* pp;
 	int gost_id;
@@ -163,9 +178,15 @@ ldns_key_new_frm_fp_gost_l(FILE* fp, int* line_nr)
 	if(!gost_id)
 		return NULL;
 
-	len = ldns_fget_token_l(fp, token, "", sizeof(token), line_nr);
-	if(len == -1)
+	if (ldns_fget_keyword_data_l(fp, "GostAsn1", ": ", token, "\n", 
+		sizeof(token), line_nr) == -1)
 		return NULL;
+	while(strlen(token) < 96) {
+		/* read more b64 from the file, b64 split on multiple lines */
+		if(ldns_fget_token_l(fp, token+strlen(token), "\n",
+			sizeof(token)-strlen(token), line_nr) == -1)
+			return NULL;
+	}
 	if(ldns_str2rdf_b64(&b64rdf, token) != LDNS_STATUS_OK)
 		return NULL;
 	pp = (unsigned char*)ldns_rdf_data(b64rdf);
@@ -226,22 +247,22 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 		alg = LDNS_SIGN_RSAMD5;
 	}
 	if (strncmp(d, "2 DH", 2) == 0) {
-		alg = LDNS_DH;
+		alg = (ldns_signing_algorithm)LDNS_DH;
 	}
 	if (strncmp(d, "3 DSA", 2) == 0) {
 		alg = LDNS_SIGN_DSA;
 	}
 	if (strncmp(d, "4 ECC", 2) == 0) {
-		alg = LDNS_ECC;
+		alg = (ldns_signing_algorithm)LDNS_ECC;
 	}
 	if (strncmp(d, "5 RSASHA1", 2) == 0) {
 		alg = LDNS_SIGN_RSASHA1;
 	}
 	if (strncmp(d, "6 DSA", 2) == 0) {
-		alg = LDNS_DSA_NSEC3;
+		alg = LDNS_SIGN_DSA_NSEC3;
 	}
 	if (strncmp(d, "7 RSASHA1", 2) == 0) {
-		alg = LDNS_RSASHA1_NSEC3;
+		alg = LDNS_SIGN_RSASHA1_NSEC3;
 	}
 
 	if (strncmp(d, "8 RSASHA256", 2) == 0) {
@@ -260,12 +281,12 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 		fprintf(stderr, "version of ldns\n");
 #endif
 	}
-	if (strncmp(d, "211 GOST", 4) == 0) {
+	if (strncmp(d, "12 ECC-GOST", 3) == 0) {
 #ifdef USE_GOST
-		alg = LDNS_SIGN_GOST;
+		alg = LDNS_SIGN_ECC_GOST;
 #else
-		fprintf(stderr, "Warning: GOST not compiled into this ");
-		fprintf(stderr, "version of ldns\n");
+		fprintf(stderr, "Warning: ECC-GOST not compiled into this ");
+		fprintf(stderr, "version of ldns, use --enable-gost\n");
 #endif
 	}
 	if (strncmp(d, "157 HMAC-MD5", 4) == 0) {
@@ -326,9 +347,13 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 			ldns_key_set_hmac_key(k, hmac);
 #endif /* HAVE_SSL */
 			break;
-		case LDNS_SIGN_GOST:
+		case LDNS_SIGN_ECC_GOST:
 			ldns_key_set_algorithm(k, alg);
 #if defined(HAVE_SSL) && defined(USE_GOST)
+                        if(!ldns_key_EVP_load_gost_id()) {
+				ldns_key_free(k);
+                                return LDNS_STATUS_CRYPTO_ALGO_NOT_IMPL;
+                        }
 			ldns_key_set_evp_key(k, 
 				ldns_key_new_frm_fp_gost_l(fp, line_nr));
 			if(!k->_key.key) {
@@ -727,7 +752,7 @@ ldns_key_new_frm_algorithm(ldns_signing_algorithm alg, uint16_t size)
 
 			ldns_key_set_flags(k, 0);
 			break;
-		case LDNS_SIGN_GOST:
+		case LDNS_SIGN_ECC_GOST:
 #if defined(HAVE_SSL) && defined(USE_GOST)
 			ldns_key_set_evp_key(k, ldns_gen_gost_key());
 #endif /* HAVE_SSL and USE_GOST */
@@ -988,7 +1013,6 @@ ldns_key_list_push_key(ldns_key_list *key_list, ldns_key *key)
         ldns_key **keys;
 
         key_count = ldns_key_list_key_count(key_list);
-        keys = key_list->_keys;
 
         /* grow the array */
         keys = LDNS_XREALLOC(
@@ -1220,7 +1244,7 @@ ldns_key2rr(const ldns_key *k)
 			}
 #endif /* HAVE_SSL */
 			break;
-		case LDNS_SIGN_GOST:
+		case LDNS_SIGN_ECC_GOST:
 			ldns_rr_push_rdf(pubkey, ldns_native2rdf_int8(
 				LDNS_RDF_TYPE_ALG, ldns_key_algorithm(k)));
 #if defined(HAVE_SSL) && defined(USE_GOST)
