@@ -9,6 +9,10 @@
 #ifndef LDNS_TSIG_H
 #define LDNS_TSIG_H
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /**
  * \file
  *
@@ -47,10 +51,14 @@ bool ldns_pkt_tsig_verify(ldns_pkt *pkt, uint8_t *wire, size_t wire_size, const
  * \param[in] key_name the name of the shared key
  * \param[in] key_data the key in base 64 format
  * \param[in] fudge seconds of error permitted in time signed
- * \param[in] algorithm_name the name of the algorithm used 
+ * \param[in] algorithm_name the name of the algorithm used
  * \param[in] query_mac is added to the digest if not NULL (so NULL is for signing queries, not NULL is for signing answers)
  * \return status (OK if success)
  */
 ldns_status ldns_pkt_tsig_sign(ldns_pkt *pkt, const char *key_name, const char *key_data, uint16_t fudge, const char *algorithm_name, ldns_rdf *query_mac);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* LDNS_TSIG_H */
