@@ -44,7 +44,6 @@ ldns_send(ldns_pkt **result_packet, ldns_resolver *r, const ldns_pkt *query_pkt)
 		tsig_mac = ldns_rr_rdf(ldns_pkt_tsig(query_pkt), 3);
 	}
 
-
 	if (!query_pkt ||
 	    ldns_pkt2buffer_wire(qb, query_pkt) != LDNS_STATUS_OK) {
 		result = LDNS_STATUS_ERR;
@@ -53,7 +52,7 @@ ldns_send(ldns_pkt **result_packet, ldns_resolver *r, const ldns_pkt *query_pkt)
 	}
 
 	ldns_buffer_free(qb);
-	
+
 	return result;
 }
 
@@ -97,7 +96,6 @@ ldns_send_buffer(ldns_pkt **result, ldns_resolver *r, ldns_buffer *qb, ldns_rdf
 			/* not reachable nameserver! */
 			continue;
 		}
-		all_servers_rtt_inf = false;
 
 		/* maybe verbosity setting?
 		printf("Sending to ");
@@ -106,17 +104,21 @@ ldns_send_buffer(ldns_pkt **result, ldns_resolver *r, ldns_buffer *qb, ldns_rdf
 		*/
 		ns = ldns_rdf2native_sockaddr_storage(ns_array[i],
 				ldns_resolver_port(r), &ns_len);
-		
-		if ((ns->ss_family == AF_INET) && 
+
+		if ((ns->ss_family == AF_INET) &&
 				(ldns_resolver_ip6(r) == LDNS_RESOLV_INET6)) {
+			/* not reachable */
 			continue;
 		}
 
 		if ((ns->ss_family == AF_INET6) &&
 				 (ldns_resolver_ip6(r) == LDNS_RESOLV_INET)) {
+			/* not reachable */
 			continue;
 		}
 
+		all_servers_rtt_inf = false;
+
 		gettimeofday(&tv_s, NULL);
 
 		send_status = LDNS_STATUS_ERR;
@@ -210,7 +212,7 @@ ldns_send_buffer(ldns_pkt **result, ldns_resolver *r, ldns_buffer *qb, ldns_rdf
 #else
 	(void)tsig_mac;
 #endif /* HAVE_SSL */
-	
+
 	LDNS_FREE(reply_bytes);
 	if (result) {
 		*result = reply;
@@ -281,7 +283,7 @@ ldns_sock_wait(int sockfd, struct timeval timeout, int write)
 }
 
 ldns_status
-ldns_udp_send(uint8_t **result, ldns_buffer *qbin, const struct sockaddr_storage *to, 
+ldns_udp_send(uint8_t **result, ldns_buffer *qbin, const struct sockaddr_storage *to,
 		socklen_t tolen, struct timeval timeout, size_t *answer_size)
 {
 	int sockfd;
@@ -298,7 +300,7 @@ ldns_udp_send(uint8_t **result, ldns_buffer *qbin, const struct sockaddr_storage
 		close(sockfd);
 		return LDNS_STATUS_NETWORK_ERR;
 	}
-	
+
 	answer = ldns_udp_read_wire(sockfd, answer_size, NULL, NULL);
 	close(sockfd);
 
@@ -333,7 +335,7 @@ int
 ldns_udp_connect(const struct sockaddr_storage *to, struct timeval ATTR_UNUSED(timeout))
 {
 	int sockfd;
-	
+
 	if ((sockfd = socket((int)((struct sockaddr*)to)->sa_family, SOCK_DGRAM, 
 					IPPROTO_UDP)) 
 			== -1) {
@@ -347,7 +349,7 @@ ldns_tcp_connect(const struct sockaddr_storage *to, socklen_t tolen,
 		struct timeval timeout)
 {
 	int sockfd;
-	
+
 	if ((sockfd = socket((int)((struct sockaddr*)to)->sa_family, SOCK_STREAM, 
 					IPPROTO_TCP)) == -1) {
 		return 0;
@@ -502,7 +504,7 @@ ldns_tcp_read_wire_timeout(int sockfd, size_t *size, struct timeval timeout)
 {
 	uint8_t *wire;
 	uint16_t wire_size;
-	ssize_t bytes = 0;
+	ssize_t bytes = 0, rc = 0;
 
 	wire = LDNS_XMALLOC(uint8_t, 2);
 	if (!wire) {
@@ -536,13 +538,14 @@ ldns_tcp_read_wire_timeout(int sockfd, size_t *size, struct timeval timeout)
 			LDNS_FREE(wire);
 			return NULL;
 		}
-		bytes += recv(sockfd, (void*) (wire + bytes), 
+		rc = recv(sockfd, (void*) (wire + bytes), 
 				(size_t) (wire_size - bytes), 0);
-		if (bytes == -1 || bytes == 0) {
+		if (rc == -1 || rc == 0) {
 			LDNS_FREE(wire);
 			*size = 0;
 			return NULL;
 		}
+                bytes += rc;
 	}
 	
 	*size = (size_t) bytes;
@@ -554,7 +557,7 @@ ldns_tcp_read_wire(int sockfd, size_t *size)
 {
 	uint8_t *wire;
 	uint16_t wire_size;
-	ssize_t bytes = 0;
+	ssize_t bytes = 0, rc = 0;
 
 	wire = LDNS_XMALLOC(uint8_t, 2);
 	if (!wire) {
@@ -578,13 +581,14 @@ ldns_tcp_read_wire(int sockfd, size_t *size)
 	bytes = 0;
 
 	while (bytes < (ssize_t) wire_size) {
-		bytes += recv(sockfd, (void*) (wire + bytes), 
+		rc = recv(sockfd, (void*) (wire + bytes), 
 				(size_t) (wire_size - bytes), 0);
-		if (bytes == -1 || bytes == 0) {
+		if (rc == -1 || rc == 0) {
 			LDNS_FREE(wire);
 			*size = 0;
 			return NULL;
 		}
+                bytes += rc;
 	}
 	
 	*size = (size_t) bytes;
@@ -734,7 +738,7 @@ ldns_axfr_start(ldns_resolver *resolver, ldns_rdf *domain, ldns_rr_class class)
                 return LDNS_STATUS_ADDRESS_ERR;
         }
         /* For AXFR, we have to make the connection ourselves */
-        /* try all nameservers (which usually would mean v4 fallback if 
+        /* try all nameservers (which usually would mean v4 fallback if
          * @hostname is used */
         for (ns_i = 0;
              ns_i < ldns_resolver_nameserver_count(resolver) &&
@@ -743,8 +747,8 @@ ldns_axfr_start(ldns_resolver *resolver, ldns_rdf *domain, ldns_rr_class class)
 	        ns = ldns_rdf2native_sockaddr_storage(
 	        	resolver->_nameservers[ns_i],
 			ldns_resolver_port(resolver), &ns_len);
-	
-		resolver->_socket = ldns_tcp_connect(ns, (socklen_t)ns_len, 
+
+		resolver->_socket = ldns_tcp_connect(ns, (socklen_t)ns_len,
 				ldns_resolver_timeout(resolver));
 	}
 
@@ -761,26 +765,45 @@ ldns_axfr_start(ldns_resolver *resolver, ldns_rdf *domain, ldns_rr_class class)
 		                            ldns_resolver_tsig_keydata(resolver),
 		                            300, ldns_resolver_tsig_algorithm(resolver), NULL);
 		if (status != LDNS_STATUS_OK) {
+			/* RoRi: to prevent problems on subsequent calls to ldns_axfr_start
+			   we have to close the socket here! */
+			close(resolver->_socket);
+			resolver->_socket = 0;
+
 			return LDNS_STATUS_CRYPTO_TSIG_ERR;
 		}
 	}
 #endif /* HAVE_SSL */
 
-        /* Convert the query to a buffer          * Is this necessary?
+        /* Convert the query to a buffer
+         * Is this necessary?
          */
         query_wire = ldns_buffer_new(LDNS_MAX_PACKETLEN);
         status = ldns_pkt2buffer_wire(query_wire, query);
         if (status != LDNS_STATUS_OK) {
                 ldns_pkt_free(query);
                 LDNS_FREE(ns);
+
+		/* RoRi: to prevent problems on subsequent calls to ldns_axfr_start
+		    we have to close the socket here! */
+		close(resolver->_socket);
+		resolver->_socket = 0;
+
                 return status;
         }
         /* Send the query */
-        if (ldns_tcp_send_query(query_wire, resolver->_socket, ns, 
+        if (ldns_tcp_send_query(query_wire, resolver->_socket, ns,
 				(socklen_t)ns_len) == 0) {
                 ldns_pkt_free(query);
                 ldns_buffer_free(query_wire);
                 LDNS_FREE(ns);
+
+		/* RoRi: to prevent problems on subsequent calls to ldns_axfr_start
+		         we have to close the socket here! */
+
+		close(resolver->_socket);
+		resolver->_socket = 0;
+
                 return LDNS_STATUS_NETWORK_ERR;
         }
 
