@@ -24,7 +24,7 @@
 
 #include <errno.h>
 #ifdef HAVE_NETDB_H
-#include <netdb.h> 
+#include <netdb.h>
 #endif
 
 #include <limits.h>
@@ -35,12 +35,12 @@
 ldns_status
 ldns_str2rdf_int16(ldns_rdf **rd, const char *shortstr)
 {
-	char *end = NULL;    
+	char *end = NULL;
 	uint16_t *r;
 	r = LDNS_MALLOC(uint16_t);
-	
+
 	*r = htons((uint16_t)strtol((char *)shortstr, &end, 10));
-	
+
 	if(*end != 0) {
 		LDNS_FREE(r);
 		return LDNS_STATUS_INVALID_INT;
@@ -81,7 +81,7 @@ ldns_str2rdf_time(ldns_rdf **rd, const char *time)
 		if (tm.tm_mday < 1 || tm.tm_mday > 31) {
 			goto bad_format;
 		}
-		
+
 		if (tm.tm_hour < 0 || tm.tm_hour > 23) {
 			goto bad_format;
 		}
@@ -114,7 +114,7 @@ ldns_str2rdf_time(ldns_rdf **rd, const char *time)
 			return LDNS_STATUS_OK;
 		}
 	}
-	
+
 	bad_format:
 	LDNS_FREE(r);
 	return LDNS_STATUS_INVALID_TIME;
@@ -139,7 +139,7 @@ ldns_str2rdf_nsec3_salt(ldns_rdf **rd, const char *salt_str)
 	if (salt_length_str > 512) {
 		return LDNS_STATUS_INVALID_HEX;
 	}
-	
+
 	salt = LDNS_XMALLOC(uint8_t, salt_length_str / 2);
 	for (c = 0; c < salt_length_str; c += 2) {
 		if (isxdigit((int) salt_str[c]) && isxdigit((int) salt_str[c+1])) {
@@ -151,7 +151,7 @@ ldns_str2rdf_nsec3_salt(ldns_rdf **rd, const char *salt_str)
 		}
 	}
 	salt_length = (uint8_t) (salt_length_str / 2);
-	
+
 	data = LDNS_XMALLOC(uint8_t, 1 + salt_length);
 	data[0] = salt_length;
 	memcpy(&data[1], salt, salt_length);
@@ -181,17 +181,19 @@ ldns_str2rdf_period(ldns_rdf **rd,const char *period)
 	return LDNS_STATUS_OK;
 }
 
-ldns_status 
+ldns_status
 ldns_str2rdf_int32(ldns_rdf **rd, const char *longstr)
 {
-	char *end;  
+	char *end;
 	uint16_t *r = NULL;
 	uint32_t l;
 
 	r = (uint16_t*)LDNS_MALLOC(uint32_t);
-	errno = 0; /* must set to zero before call, 
+	errno = 0; /* must set to zero before call,
 			note race condition on errno */
-	l = htonl((uint32_t)strtol((char*)longstr, &end, 10));
+	if(*longstr == '-')
+		l = htonl((uint32_t)strtol((char*)longstr, &end, 10));
+	else	l = htonl((uint32_t)strtoul((char*)longstr, &end, 10));
 
 	if(*end != 0) {
 		LDNS_FREE(r);
@@ -212,11 +214,11 @@ ldns_str2rdf_int32(ldns_rdf **rd, const char *longstr)
 ldns_status
 ldns_str2rdf_int8(ldns_rdf **rd, const char *bytestr)
 {
-	char *end;     
+	char *end;
 	uint8_t *r = NULL;
 
 	r = LDNS_MALLOC(uint8_t);
- 
+
 	*r = (uint8_t)strtol((char*)bytestr, &end, 10);
 
         if(*end != 0) {
@@ -237,7 +239,7 @@ ldns_str2rdf_int8(ldns_rdf **rd, const char *bytestr)
  *
  * The string pointer at *s is increased by either 0 (on error), 1 (on
  * normal escapes), or 3 (on octals)
- * 
+ *
  * Returns the number of bytes read from the escaped string, or
  * 0 on error
  */
@@ -256,9 +258,9 @@ parse_escape(uint8_t *s, uint8_t *q) {
 		return 3;
 	} else {
 		s++;
-		if (*s == '\0') {
+		if (*s == '\0' || isdigit((int) *s)) {
 			/* apparently the string terminator
-			 * has been escaped...
+			 * or a digit has been escaped...
 		         */
 			return 0;
 		}
@@ -278,10 +280,10 @@ ldns_str2rdf_dname(ldns_rdf **d, const char *str)
 	size_t len;
 
 	int esc;
-	uint8_t *s,*p,*q, *pq, label_len;
+	uint8_t *s, *q, *pq, label_len;
 	uint8_t buf[LDNS_MAX_DOMAINLEN + 1];
 	*d = NULL;
-	
+
 	len = strlen((char*)str);
 	/* octet representation can make strings a lot longer than actual length */
 	if (len > LDNS_MAX_DOMAINLEN * 4) {
@@ -289,28 +291,29 @@ ldns_str2rdf_dname(ldns_rdf **d, const char *str)
 	}
 	if (0 == len) {
 		return LDNS_STATUS_DOMAINNAME_UNDERFLOW;
-	} 
-	
+	}
+
 	/* root label */
 	if (1 == len && *str == '.') {
-		*d = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_DNAME, 1, "\0"); 
+		*d = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_DNAME, 1, "\0");
 		return LDNS_STATUS_OK;
 	}
 
 	/* get on with the rest */
 
-	/* s is on the current dot
-	 * p on the previous one
-	 * q builds the dname
+	/* s is on the current character in the string
+         * pq points to where the labellength is going to go
+         * label_len keeps track of the current label's length
+	 * q builds the dname inside the buf array
 	 */
 	len = 0;
 	q = buf+1;
 	pq = buf;
 	label_len = 0;
-	for (s = p = (uint8_t *) str; *s; s++, q++) {
+	for (s = (uint8_t *)str; *s; s++, q++) {
 		if (q > buf + LDNS_MAX_DOMAINLEN) {
 			return LDNS_STATUS_DOMAINNAME_OVERFLOW;
-		}	
+		}
 		*q = 0;
 		switch (*s) {
 		case '.':
@@ -324,7 +327,6 @@ ldns_str2rdf_dname(ldns_rdf **d, const char *str)
 			*pq = label_len;
 			label_len = 0;
 			pq = q;
-			p = s+1;
 			break;
 		case '\\':
 			/* octet value or literal char */
@@ -347,13 +349,19 @@ ldns_str2rdf_dname(ldns_rdf **d, const char *str)
 		if (q > buf + LDNS_MAX_DOMAINLEN) {
 			return LDNS_STATUS_DOMAINNAME_OVERFLOW;
 		}
+                if (label_len > LDNS_MAX_LABELLEN) {
+                        return LDNS_STATUS_LABEL_OVERFLOW;
+                }
+                if (label_len == 0) { /* label_len 0 but not . at end? */
+                        return LDNS_STATUS_EMPTY_LABEL;
+                }
 		len += label_len + 1;
 		*pq = label_len;
 		*q = 0;
 	}
 	len++;
 
-	*d = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_DNAME, len, buf); 
+	*d = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_DNAME, len, buf);
 	return LDNS_STATUS_OK;
 }
 
@@ -388,7 +396,7 @@ ldns_status
 ldns_str2rdf_str(ldns_rdf **rd, const char *str)
 {
 	uint8_t *data;
-	size_t i, str_i;
+	size_t i, str_i, esc_i;
 
 	if (strlen(str) > 255) {
 		return LDNS_STATUS_INVALID_STR;
@@ -397,7 +405,17 @@ ldns_str2rdf_str(ldns_rdf **rd, const char *str)
 	data = LDNS_XMALLOC(uint8_t, strlen(str) + 1);
 	i = 1;
 	for (str_i = 0; str_i < strlen(str); str_i++) {
-		data[i] = (uint8_t) str[str_i];
+		if (str[str_i] == '\\') {
+			/* octet value or literal char */
+			esc_i = (size_t) parse_escape((uint8_t*) &str[str_i], (uint8_t*) &data[i]);
+			if (esc_i == 0) {
+				LDNS_FREE(data);
+				return LDNS_STATUS_SYNTAX_BAD_ESCAPE;
+			}
+			str_i += esc_i;
+		} else {
+			data[i] = (uint8_t) str[str_i];
+		}
 		i++;
 	}
 	data[0] = i - 1;
@@ -421,7 +439,7 @@ ldns_str2rdf_apl(ldns_rdf **rd, const char *str)
 	uint8_t prefix;
 
 	uint8_t *data;
-	
+
 	size_t i = 0;
 
 	/* [!]afi:address/prefix */
@@ -437,7 +455,7 @@ ldns_str2rdf_apl(ldns_rdf **rd, const char *str)
 	}
 
 	family = (uint16_t) atoi(my_str);
-	
+
 	my_str = strchr(my_str, ':') + 1;
 
 	/* need ip addr and only ip addr for inet_pton */
@@ -485,7 +503,7 @@ ldns_str2rdf_apl(ldns_rdf **rd, const char *str)
 		/* set bit 1 of byte 3 */
 		data[3] = data[3] | 0x80;
 	}
-	
+
 	memcpy(data + 4, afdpart, afdlength);
 
 	*rd = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_APL, afdlength + 4, data);
@@ -501,10 +519,10 @@ ldns_str2rdf_b64(ldns_rdf **rd, const char *str)
 {
 	uint8_t *buffer;
 	int16_t i;
-	
+
 	buffer = LDNS_XMALLOC(uint8_t, ldns_b64_ntop_calculate_size(strlen(str)));
-	
-	i = (uint16_t)ldns_b64_pton((const char*)str, buffer, 
+
+	i = (uint16_t)ldns_b64_pton((const char*)str, buffer,
 						   ldns_b64_ntop_calculate_size(strlen(str)));
 	if (-1 == i) {
 		LDNS_FREE(buffer);
@@ -527,8 +545,8 @@ ldns_str2rdf_b32_ext(ldns_rdf **rd, const char *str)
 	uint8_t len = ldns_b32_pton_calculate_size(strlen(str));
 	buffer = LDNS_XMALLOC(uint8_t, len + 1);
 	buffer[0] = len;
-	
-	i = ldns_b32_pton_extended_hex((const char*)str, strlen(str), buffer + 1, 
+
+	i = ldns_b32_pton_extended_hex((const char*)str, strlen(str), buffer + 1,
 							 ldns_b32_ntop_calculate_size(strlen(str)));
 	if (i < 0) {
 		return LDNS_STATUS_INVALID_B32_EXT;
@@ -544,19 +562,19 @@ ldns_str2rdf_b32_ext(ldns_rdf **rd, const char *str)
 ldns_status
 ldns_str2rdf_hex(ldns_rdf **rd, const char *str)
 {
-        uint8_t *t, *t_orig;
-        int i;
+	uint8_t *t, *t_orig;
+	int i;
 	size_t len;
 
 	len = strlen(str);
 
 	if (len > LDNS_MAX_RDFLEN * 2) {
 		return LDNS_STATUS_LABEL_OVERFLOW;
-        } else {
+	} else {
 		t = LDNS_XMALLOC(uint8_t, (len / 2) + 1);
 		t_orig = t;
-                /* Now process octet by octet... */
-                while (*str) {
+		/* Now process octet by octet... */
+		while (*str) {
 			*t = 0;
 			if (isspace((int) *str)) {
 				str++;
@@ -574,13 +592,13 @@ ldns_str2rdf_hex(ldns_rdf **rd, const char *str)
 				}
 				++t;
 			}
-                }
+		}
 		*rd = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_HEX,
 		                            (size_t) (t - t_orig),
 		                            t_orig);
 		LDNS_FREE(t_orig);
-        }
-        return LDNS_STATUS_OK;
+	}
+	return LDNS_STATUS_OK;
 }
 
 ldns_status
@@ -597,12 +615,12 @@ ldns_str2rdf_nsec(ldns_rdf **rd, const char *str)
 	str_buf = LDNS_MALLOC(ldns_buffer);
 	ldns_buffer_new_frm_data(str_buf, (char *)str, strlen(str));
 
-	while ((c = ldns_bget_token(str_buf, token, delimiters, LDNS_MAX_RDFLEN)) != -1) {
+	while ((c = ldns_bget_token(str_buf, token, delimiters, LDNS_MAX_RDFLEN)) != -1 && c != 0) {
 		cur_type = ldns_get_rr_type_by_name(token);
 		type_list[type_count] = cur_type;
 		type_count++;
 	}
-	
+
 	*rd = ldns_dnssec_create_nsec_bitmap(type_list,
 	                                     type_count,
 	                                     LDNS_RR_TYPE_NSEC);
@@ -665,7 +683,7 @@ ldns_str2rdf_cert_alg(ldns_rdf **rd, const char *str)
 
 	return st;
 }
-		
+
 /* An alg field can either be specified as a 8 bits number
  * or by its symbolic name. Handle both
  */
@@ -690,11 +708,11 @@ ldns_str2rdf_alg(ldns_rdf **rd, const char *str)
 	}
 	return st;
 }
-		
+
 ldns_status
 ldns_str2rdf_unknown(ldns_rdf **rd, const char *str)
 {
-	/* this should be caught in an earlier time (general str2host for 
+	/* this should be caught in an earlier time (general str2host for
 	   rr's */
 	rd = rd;
 	str = str;
@@ -745,7 +763,7 @@ loc_parse_cm(char* my_str, char** endstr, uint8_t* m, uint8_t* e)
 		val /= 10;
 	}
 	*m = (uint8_t)val;
-	
+
 	if (*e > 9)
 		return 0;
 	if (*my_str == 'm' || *my_str == 'M') {
@@ -770,7 +788,7 @@ ldns_str2rdf_loc(ldns_rdf **rd, const char *str)
 	uint8_t size_b = 1, size_e = 2;
 	uint8_t horiz_pre_b = 1, horiz_pre_e = 6;
 	uint8_t vert_pre_b = 1, vert_pre_e = 3;
-	
+
 	double s = 0.0;
 	bool northerness;
 	bool easterness;
@@ -933,16 +951,16 @@ ldns_str2rdf_wks(ldns_rdf **rd, const char *str)
 	uint8_t *bitmap = NULL;
 	uint8_t *data;
 	int bm_len = 0;
-	
+
 	struct protoent *proto = NULL;
 	struct servent *serv = NULL;
 	int serv_port;
-	
+
 	ldns_buffer *str_buf;
-	
+
 	char *proto_str = NULL;
 	char *token = LDNS_XMALLOC(char, 50);
-	
+
 	str_buf = LDNS_MALLOC(ldns_buffer);
 	ldns_buffer_new_frm_data(str_buf, (char *)str, strlen(str));
 
@@ -951,7 +969,7 @@ ldns_str2rdf_wks(ldns_rdf **rd, const char *str)
 			proto_str = strdup(token);
 			if (!proto_str) {
 				LDNS_FREE(token);
-				LDNS_FREE(str_buf);
+	                        ldns_buffer_free(str_buf);
 				return LDNS_STATUS_INVALID_STR;
 			}
 		} else {
@@ -961,8 +979,14 @@ ldns_str2rdf_wks(ldns_rdf **rd, const char *str)
 			} else {
 				serv_port = atoi(token);
 			}
-			if (serv_port / 8 > bm_len) {
+			if (serv_port / 8 >= bm_len) {
 				bitmap = LDNS_XREALLOC(bitmap, uint8_t, (serv_port / 8) + 1);
+                                if(!bitmap) {
+				        LDNS_FREE(token);
+	                                ldns_buffer_free(str_buf);
+				        free(proto_str);
+				        return LDNS_STATUS_INVALID_STR;
+                                }
 				/* set to zero to be sure */
 				for (; bm_len <= serv_port / 8; bm_len++) {
 					bitmap[bm_len] = 0;
@@ -971,22 +995,44 @@ ldns_str2rdf_wks(ldns_rdf **rd, const char *str)
 			ldns_set_bit(bitmap + (serv_port / 8), 7 - (serv_port % 8), true);
 		}
 	}
-	
+
+	if (!proto_str) {
+		LDNS_FREE(token);
+	        ldns_buffer_free(str_buf);
+		return LDNS_STATUS_INVALID_STR;
+	}
+        if(!bitmap) {
+	        LDNS_FREE(token);
+	        ldns_buffer_free(str_buf);
+	        free(proto_str);
+	        return LDNS_STATUS_INVALID_STR;
+        }
+
 	data = LDNS_XMALLOC(uint8_t, bm_len + 1);
-	proto = getprotobyname(proto_str);
+        if(!data) {
+	        LDNS_FREE(token);
+	        ldns_buffer_free(str_buf);
+	        LDNS_FREE(bitmap);
+	        free(proto_str);
+	        return LDNS_STATUS_INVALID_STR;
+        }
+    if (proto_str)
+		proto = getprotobyname(proto_str);
 	if (proto) {
 		data[0] = (uint8_t) proto->p_proto;
-	} else {
+	} else if (proto_str) {
 		data[0] = (uint8_t) atoi(proto_str);
+	} else {
+		data[0] = 0;
 	}
 	memcpy(data + 1, bitmap, (size_t) bm_len);
-	
+
 	*rd = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_WKS, (uint16_t) (bm_len + 1), data);
 
+	LDNS_FREE(data);
 	LDNS_FREE(token);
 	ldns_buffer_free(str_buf);
 	LDNS_FREE(bitmap);
-	LDNS_FREE(data);
 	free(proto_str);
 #ifdef HAVE_ENDSERVENT
 	endservent();
@@ -994,17 +1040,173 @@ ldns_str2rdf_wks(ldns_rdf **rd, const char *str)
 #ifdef HAVE_ENDPROTOENT
 	endprotoent();
 #endif
-	
+
 	return LDNS_STATUS_OK;
 }
 
 ldns_status
 ldns_str2rdf_nsap(ldns_rdf **rd, const char *str)
 {
+    size_t len, i;
+    char* nsap_str = (char*) str;
+
 	/* just a hex string with optional dots? */
 	if (str[0] != '0' || str[1] != 'x') {
 		return LDNS_STATUS_INVALID_STR;
 	} else {
+		len = strlen(str);
+		for (i=0; i < len; i++) {
+			if (nsap_str[i] == '.')
+				nsap_str[i] = ' ';
+        }
 		return ldns_str2rdf_hex(rd, str+2);
 	}
 }
+
+ldns_status
+ldns_str2rdf_atma(ldns_rdf **rd, const char *str)
+{
+    size_t len, i;
+    char* atma_str = (char*) str;
+	ldns_status status;
+
+	/* just a hex string with optional dots? */
+	len = strlen(str);
+	for (i=0; i < len; i++) {
+		if (atma_str[i] == '.')
+			atma_str[i] = ' ';
+	}
+	status = ldns_str2rdf_hex(rd, str);
+    if (status != LDNS_STATUS_OK) {
+		; /* probably in e.164 format than */
+	}
+	return status;
+}
+
+ldns_status
+ldns_str2rdf_ipseckey(ldns_rdf **rd, const char *str)
+{
+	uint8_t precedence = 0;
+	uint8_t gateway_type = 0;
+	uint8_t algorithm = 0;
+	char* gateway = NULL;
+	char* publickey = NULL;
+	uint8_t *data;
+	ldns_buffer *str_buf;
+	char *token = LDNS_XMALLOC(char, 256);
+	int token_count = 0;
+	int ipseckey_len = 0;
+	ldns_rdf* gateway_rdf = NULL;
+	ldns_rdf* publickey_rdf = NULL;
+	ldns_status status = LDNS_STATUS_OK;
+
+	str_buf = LDNS_MALLOC(ldns_buffer);
+	ldns_buffer_new_frm_data(str_buf, (char *)str, strlen(str));
+	while(ldns_bget_token(str_buf, token, "\t\n ", strlen(str)) > 0) {
+		switch (token_count) {
+				case 0:
+					precedence = atoi(token);
+					break;
+				case 1:
+					gateway_type = atoi(token);
+					break;
+				case 2:
+					algorithm = atoi(token);
+					break;
+				case 3:
+					gateway = strdup(token);
+					if (!gateway || (gateway_type == 0 &&
+							(token[0] != '.' || token[1] != '\0'))) {
+						LDNS_FREE(gateway);
+						LDNS_FREE(token);
+						LDNS_FREE(str_buf);
+						return LDNS_STATUS_INVALID_STR;
+					}
+					break;
+				case 4:
+					publickey = strdup(token);
+					break;
+				default:
+					LDNS_FREE(token);
+					LDNS_FREE(str_buf);
+					return LDNS_STATUS_INVALID_STR;
+					break;
+		}
+		token_count++;
+	}
+
+	if (!gateway || !publickey) {
+		if (gateway)
+			LDNS_FREE(gateway);
+		if (publickey)
+			LDNS_FREE(publickey);
+		LDNS_FREE(token);
+		LDNS_FREE(str_buf);
+		return LDNS_STATUS_INVALID_STR;
+	}
+
+	if (gateway_type == 1) {
+		status = ldns_str2rdf_a(&gateway_rdf, gateway);
+	} else if (gateway_type == 2) {
+		status = ldns_str2rdf_aaaa(&gateway_rdf, gateway);
+	} else if (gateway_type == 3) {
+		status = ldns_str2rdf_dname(&gateway_rdf, gateway);
+	}
+
+	if (status != LDNS_STATUS_OK) {
+		if (gateway)
+			LDNS_FREE(gateway);
+		if (publickey)
+			LDNS_FREE(publickey);
+		LDNS_FREE(token);
+		LDNS_FREE(str_buf);
+		return LDNS_STATUS_INVALID_STR;
+	}
+
+	status = ldns_str2rdf_b64(&publickey_rdf, publickey);
+
+	if (status != LDNS_STATUS_OK) {
+		if (gateway)
+			LDNS_FREE(gateway);
+		if (publickey)
+			LDNS_FREE(publickey);
+		LDNS_FREE(token);
+		LDNS_FREE(str_buf);
+		return LDNS_STATUS_INVALID_STR;
+	}
+
+	/* now copy all into one ipseckey rdf */
+	if (gateway_type)
+		ipseckey_len = 3 + ldns_rdf_size(gateway_rdf) + ldns_rdf_size(publickey_rdf);
+	else
+		ipseckey_len = 3 + ldns_rdf_size(publickey_rdf);
+
+	data = LDNS_XMALLOC(uint8_t, ipseckey_len);
+
+	data[0] = precedence;
+	data[1] = gateway_type;
+	data[2] = algorithm;
+
+	if (gateway_type) {
+		memcpy(data + 3,
+			ldns_rdf_data(gateway_rdf), ldns_rdf_size(gateway_rdf));
+		memcpy(data + 3 + ldns_rdf_size(gateway_rdf),
+			ldns_rdf_data(publickey_rdf), ldns_rdf_size(publickey_rdf));
+	} else {
+		memcpy(data + 3,
+			ldns_rdf_data(publickey_rdf), ldns_rdf_size(publickey_rdf));
+	}
+
+	*rd = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_IPSECKEY, (uint16_t) ipseckey_len, data);
+
+	if (gateway)
+		LDNS_FREE(gateway);
+	if (publickey)
+		LDNS_FREE(publickey);
+	LDNS_FREE(token);
+	ldns_buffer_free(str_buf);
+	ldns_rdf_free(gateway_rdf);
+	ldns_rdf_free(publickey_rdf);
+	LDNS_FREE(data);
+	return LDNS_STATUS_OK;
+}
